/*****************************************************************************
 *
 *                        CalC (Calcium Calculator)
 *                             Victor Matveev
 *
 *                              simulation.cpp
 *
 *  Defines the main highest-level simulation compound object SimulationObj, 
 *  which contains the concentration fields (FieldObj "Ca" object and the 
 *  BufferArray object) and the remaining scalar time-dependent variables 
 *  combined in the KineticObj variable "Gates". Also links the grid object 
 *  (GridObj *Grid), the geometry object (RegionObj *Synapse), the plot array
 *  variable (PlotArray *Plots), and the parsed script (TokenString *Params).
 *
 *  Includes as members the highest-level adaptive "Adaptive(double)" and 
 *  non-adaptive "FixedTimeStep(double)" time-update routines. 
 *
 *  "Ca" on one side, and the buffers plus KineticObj variables on the other
 *  side, are updated on temporal grids staggered by dt/2
 *
 ******************************************************************************/

#include "stdafx.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>  // for compatibility with Visual C++
#include "Ygl.h"
#include <string.h>
#include <stdarg.h>
#include "vector.h"
#include "syntax.h"
#include "box.h"
#include "grid.h"
#include "field.h"
#include "table.h"
#include "peak.h"
#include "interpol.h"
#include "simulation.h"
#include "gate.h"
#include "fplot.h"

extern double CHARGE_LOSS;

extern double get_sim_time(TokenString &TS);
extern void setFieldByFunction(TokenString &TS, long p, double *Diff, const char *errStr);
extern const char *getMethod(char *smethod, CaMethod &, BufMethod &);
extern void getRun( TokenString &TS, int i, bool *adaptive=0, char *PDEmethod=0, double *time=0, 
                    double *dt=0, double *accuracy=0, double *dtStretch=0, double *ODEaccuracy=0);

//*******************************************************************************************

SimulationObj::SimulationObj(TokenString &TS) {

  initialize();

  long pos;
  char temp[512];

  totalSimTime = get_sim_time(TS); // make sure simulation time is computed before Plot Array is constructed!

  if (TS.assert("geometry","spherical") )         GEOMETRY = SPHERICAL; 
  else if (TS.assert("geometry","conical") )      GEOMETRY = CONICAL; 
  else if (TS.assert("geometry","cylindrical") )  GEOMETRY = CYLINDRICAL;
  else if (TS.assert("geometry","disc") )         GEOMETRY = DISC;
  else if (TS.assert("geometry","cartesian.1D") ) GEOMETRY = CARTESIAN1D;
  else if (TS.assert("geometry","cartesian.2D") ) GEOMETRY = CARTESIAN2D;
  else if (TS.assert("geometry","cartesian.3D") ) ;
  else if (TS.token_count("geometry"))  TS.errorMessage( TS.token_index("geometry") + 1, 0, "Unknown geometry specifier");
  else GEOMETRY = CARTESIAN3D;
  DIMENSIONALITY =  GEOMETRY & DIMENSION_MASK;
 

  Params  = &TS;
  BCArray = new BCarrayObj(TS);
  Synapse = new RegionObj(TS);
  Grid    = new GridObj( *Synapse, TS);
  Synapse->bindToGrid( *Grid );
  FieldObj::setStaticData(*Synapse, *Grid, *BCArray); 
  Ca      = new FieldObj(TS);
  if (!Ca->source_num) throw StrCpy("No calcium channels specified");
  Buffers = new BufferArray(TS);
  initTortuosity();
  Gates   = new KineticObj(this);
  if ( TS.token_count("Import", &pos) ) Import( TS.line_string( pos + 1, temp ) );
  Plots   = new PlotArray(*this);
  Plots->draw_all();
  strcpy(m_PDEscheme, "default");

  // have to initialize run parameters before the check loop, to check the bounds on parameter values
  // also have to re-initialize after the loop, to restore the default settings

  m_dt0 = 1.0e-3; m_ODEaccuracy = m_accuracy = 1e-4; m_dtStretch = 1.05; 

  for (int i=1; i <= TS.token_count("Run"); i++) // make sure all "run" and "current" statements are Ok
  try { 
     bool adaptive; double t;
     getRun( *Params, i, &adaptive, m_PDEscheme, &t, &m_dt0, &m_accuracy, &m_dtStretch, &m_ODEaccuracy);   
     getMethod(m_PDEscheme, CaStep, BufStep );
     pos = TS.token_index("current", i) + 1;
     if (TS.equal(pos, VAR_TOKEN) || TS.equal(pos, ASSIGN_TOKEN) ) pos ++;
     ExpressionObj(TS, pos, "Bad expression for the current", this, 0, &t, "t").Evaluate();
  } catch (char *str) {
     TS.errorMessage( TS.token_index("Run", i), str, "Cannot process this \"Run\"/\"current\" statement");
  }

  m_dt0 = 1.0e-3; m_ODEaccuracy = m_accuracy = 1e-4; m_dtStretch = 1.05;
  m_maxSteps = 30; m_steps = 3;

  strcpy(m_PDEscheme, "default");

  Params->get_param("adaptive.dt0", &m_dt0);             Params->get_param("adaptive.accuracy", &m_accuracy); 
  Params->get_param("adaptive.dtStretch", &m_dtStretch); 
  Params->get_int_param("adaptive.steps", &m_steps);     Params->get_int_param("adaptive.maxSteps", &m_maxSteps); 
  Params->get_param("ODE.accuracy", &m_ODEaccuracy);     Params->get_string_param("PDE.scheme", m_PDEscheme);

 }

//*******************************************************************************************

SimulationObj::~SimulationObj() {

  if (Plots)          delete Plots;
  if (Gates)          delete Gates;
  if (DiffArray)      killTortuosity();
  FieldObj::kill_tridiag();      
  if (Buffers)        delete Buffers;
  if (Ca)             delete Ca;
  if (Grid)           delete Grid;
  if (Synapse)        delete Synapse;
  if (BCArray)        delete BCArray;
}

//********************************************************************************************
//********************************************************************************************


void SimulationObj::killTortuosity() {
  if (DiffArray) {
     for (int ti = 0; ti < DiffNum; ti ++) delete [] DiffArray[ti]; 
     delete [] DiffArray;
  }
}

//*****************************************************************************

void SimulationObj::initTortuosity() {

  bool undefined = false;
  int  bi;
  double *Diff = 0;
  long l = 0;

  DiffNum = 0;
  if ( Ca->tortDefined ) DiffNum++; else undefined = true;

  for (bi = 0; bi < Buffers->buf_num; bi++) 
    if ( Buffers->array[bi]->tortDefined ) DiffNum++;
    else undefined = true;

  if ( undefined ) DiffNum ++;

  DiffArray = new double *[DiffNum];
  int ind = 0;

  if ( undefined ) {   // at least one field has undefined tortuosity
    Diff = new double[ FieldObj::Size ];
    //fprintf(stdlog,"\n Diff -> %ld\n", Diff);
    DiffArray[ind++] = Diff; 
    if ( !Params->token_count("tortuosity") && !Params->token_count("all.tortuosity") ) 
      for (l = 0; l < FieldObj::Size; l++) Diff[l] = 1.0;
    else {
      if (VERBOSE) fprintf(stdlog,"\n### Global (default) tortuosity function specified:\n");
      long p = 0;
      if ( !Params->token_count("tortuosity", &p) ) Params->token_count("all.tortuosity", &p);
      if (VERBOSE) fprintf(stdlog,"\n#### Tortuosity function = ");
      setFieldByFunction(*Params, p+1, Diff, "Bad tortuosity expression");
    }
  }

  if ( Ca->tortDefined ) DiffArray[ind++] = Ca->Diff; else Ca->Diff = Diff;
  for (bi = 0; bi < Buffers->buf_num; bi++) 
    if ( Buffers->array[bi]->tortDefined ) DiffArray[ind++] = Buffers->array[bi]->Diff; 
       else Buffers->array[bi]->Diff = Diff;

  // fprintf(stdlog,"\n ind=%d defined=%d \n", ind, defined);
}
//*****************************************************************************

double *SimulationObj::ResolveID(const char *svar, double **t) {

  // if ( isFloat(svar) ) throw makeMessage("Expected a variable name, not a constant \"%s\"",svar);

  if (VERBOSE > 8) fprintf(stdlog, ">> ResolveID: Resolving %s\n", svar);

    if (Ca) {
      //     if ( equal(svar, Ca->ID) )  
      // { if (t) *t = &Ca->Time; return (*Ca)(); }
     if ( equal(svar, "ICa") )  
       { if (t) *t = &Ca->Time; return &(Ca->ICa); }
     if ( equal(svar, "Charge.loss") )  
       { if (t) *t = &Ca->Time; return &(CHARGE_LOSS); }
    }     

    //    if (Buffers)
    //      for (j = 0; j < Buffers->buf_num; j++)
    //	if ( equal(svar,Buffers->array[j]->ID) )  
    //          { if (t) *t = &Buffers->array[0]->Time;  return (*Buffers->array[j])(); }

    if (Gates) return Gates->ResolveID(svar, t);

    return 0;
}

//**************************************************************************


const char *SimulationObj::ResolvePtr(double *ptr) {

    if (ptr == &CHARGE_LOSS) return "Charge.loss";

    if (Ca) {
       if ( ptr == Ca->elem )  return Ca->ID; 
       else if ( ptr == &(Ca->ICa) )  return "ICa";
       else if ( ptr == &CHARGE_LOSS )  return "Charge.loss";
    } 

    if (Buffers)
      for (int j = 0; j < Buffers->buf_num; j++)
	if ( ptr == Buffers->array[j]->elem )  return  Buffers->array[j]->ID; 

    if (Gates) return Gates->ResolvePtr(ptr);
    return 0;
    //fprintf(stdlog,"ResolvePtr: cannot resolve pointer %X\n", ptr );
    //exit(1);
}

//***************************************************************************

FieldObj *SimulationObj::ResolveField(const char *label) {
      FieldObj *field = 0;
      if (Ca) {  if ( equal(label, Ca->ID) ) field = Ca; }
      if (Buffers) for (int i=0; i < Buffers->buf_num; i++)
        if ( equal(label, (Buffers->array[i])->ID) ) field = Buffers->array[i];
      return field;
}


//**************************************************************************************************


void SimulationObj::Export(const char *filename)
{
  FILE *f;
  char endtag[7];
  strcpy(endtag,"endtag");
  long wrote, toWrite;

  if (VERBOSE) 
    fprintf(stdlog,"\n#### Dumping all state variables into file %s\n", filename);

  f = fopenAssure(filename, "wb", "Simulation state export", "");

 
  if (Ca) {
	toWrite = Ca->size;
    fwrite( (void *)&toWrite, sizeof(long), 1, f);
    wrote = fwrite( (void *)(Ca->elem), sizeof(double), toWrite, f);
	if ( wrote != toWrite ) 
	  throw makeMessage("Failure to export Ca profile to file %s: wrote only %ld of total %ld elements",
	                    filename, wrote, toWrite);
  }


  if (Buffers) 
	for (int i=0; i< Buffers->buf_num; i++) {
      toWrite = Buffers->array[i]->size;
      wrote = fwrite( (void *)(Buffers->array[i]->elem), sizeof(double), toWrite, f);
  	  if ( wrote != Buffers->array[i]->size ) 
 	    throw makeMessage("Failure to export %s profile to file %s: wrote only %ld of total %ld elements",
	                      Buffers->array[i]->ID, filename, wrote, toWrite);
	}

  if (Gates) {
     int m = Gates->var_num;
     fwrite( &m, sizeof(int), 1, f );
     fwrite( (void *)(Gates->var->elem), sizeof(double), Gates->var_num, f);
  }

  fwrite( (void *)endtag, sizeof(char), strlen(endtag)+1, f);
  fclose(f);
  fflush(f);
}


//**************************************************************************************************

void SimulationObj::Import(const char *filename)
{
  FILE *f;
  char endtag[7];
 
  if (VERBOSE) 
    fprintf(stdlog,"\n#### Importing all state variables from file %s\n", filename);

  f = fopenAssure(filename, "rb", "Simulation state import", "");

  long n, Size, imported;
  
  if (Ca) {
    Size = Ca->size;
    fread( (void *)&n, sizeof(long), 1, f);
    if (n != Size) 
      throw makeMessage("Can't import the concentration fields: wrong number of elements (%ld vs %ld)\n", n, Size);
	imported = fread( (void *)(Ca->elem), sizeof(double), Size, f );
	if ( imported != Size ) 
		throw makeMessage("Failure to import Ca profile from file %s: read only %ld of the total %ld elements",
		                   filename, imported, Size);
    // Ca->initialAve = Ca->average();
  }

 
  if (Buffers) 
    for (int i=0; i< Buffers->buf_num; i++) {
      imported = fread( (void *)(Buffers->array[i]->elem), sizeof(double), Size, f);
  	  if ( imported != Size ) 
	 	 throw makeMessage("Failure to import %s profile from file %s: read only %ld of the total %ld elements",
		                   Buffers->array[i]->ID, filename, imported, Size);

      // Buffers->array[i]->initialAve = Buffers->array[i]->average();
    }

  if (Gates) {
    int m = 0;
    fread( &m, sizeof(int), 1, f);
    if (m != Gates->var_num)  
      throw makeMessage("Can't import the dynamic variables: wrong number of ODE variables (%d vs %d)\n", m, Gates->var_num);
    fread( (void *)(Gates->var->elem), sizeof(double), Gates->var_num, f);
    Gates->locations->reset();
    Gates->Evaluate();
  }

  fread( (void *)endtag, sizeof(char), strlen("endtag")+1, f );
  if ( !equal(endtag,"endtag") )  
    throw makeMessage("in Import: file size mismatch (endtag=%s)\n", endtag);

  fclose(f);
}

//**************************************************************************
//          H I N E S / N O N - A D A P T I V E    M E T H O D                 
//**************************************************************************

void SimulationObj::FixedTimeStep(double T, int n)
{
  VectorObj Next(FieldObj::Size);

  double dt = (T - Ca->Time) / double(n + 0.5);

  RunStatusString *rss = 0;
  if (VERBOSE > 4) rss = new RunStatusString(30,"time",&(Ca->Time),T);						       
  else if (VERBOSE > 1) rss = new RunStatusString(1,"time",&(Ca->Time),T);						       
  Ca->Step(CaStep, Next, *Buffers, 0.5 * dt, T);  // half a step for staggering
  
  for (int i = 1; i <= n; i++)
    {
    Buffers->Step(BufStep, Next, *Ca, dt, T);
    if (VERBOSE > 4) Gates->RungeKuttaAdaptive(dt, m_ODEaccuracy, 0, rss, T);
      else Gates->RungeKuttaAdaptive(dt, m_ODEaccuracy, 0, 0, T);
    Ca->Step(CaStep, Next, *Buffers, dt, T);
    if (VERBOSE > 1)  rss->update('.');
    Plots->draw_all(); 
    }
  
  Buffers->Step(BufStep, Next, *Ca, 0.5 * dt, T); // do half a time step for staggering
  Gates->RungeKuttaAdaptive(0.5 * dt, m_ODEaccuracy, 0, 0, T);
  Plots->draw_all();

  if (rss) delete rss;
}


//**************************************************************************
//                 A D A P T I V E   M E T H O D                 
//**************************************************************************
// Iterations begin with buffers update, so
// be sure to pass Ca field iterated over dt/2 !!
//**************************************************************************

void SimulationObj::Adaptive(double T)  {
  VectorObj Next(FieldObj::Size); 
  FieldObj  Next2(*Ca), oldCa(*Ca);
  BufferArray oldBuf(*Buffers);

  double    one, two, error, buf_dt, old_dt, dt = m_dt0;
  int       rvalue = 0, flag=0;
  long between_checks = 0, total_steps = 0, since_last_divide = 0, total_backsteps = 0;

  if (m_ODEaccuracy > 1e-3) m_ODEaccuracy = 1e-3;

  RunStatusString *ODEstatus = 0, *status = 0;

  if (VERBOSE > 2) status = new RunStatusString(40,"time",&(Ca->Time),T, "dt", &dt);
  else if (VERBOSE) status = new RunStatusString(1,"time",&(Ca->Time),T);	
  if (VERBOSE > 3) ODEstatus = status;


  oldCa = *Ca;        oldBuf = *Buffers;   // store variables to recover 
  Gates->saveState(); old_dt = dt;         // when m_maxError is exceeded

  while(1) {  // ******************************************************************

    unsigned long i=0;
    if ( between_checks < 0 ) between_checks=0;
    else if ( between_checks > m_maxSteps ) { between_checks = m_maxSteps; }
    
    while (++i <= between_checks) { // loop over n="between_checks" iterations
      dt *= m_dtStretch;
      buf_dt = Ca->Time - Gates->Time + 0.5 * dt;
      Buffers->Step(BufStep, Next, *Ca, buf_dt, T); 
      Gates->RungeKuttaAdaptive(buf_dt, m_ODEaccuracy, 0, ODEstatus, T);
      rvalue = Ca->Step(CaStep, Next, *Buffers, dt, T);
      Plots->draw_all();
      total_steps ++; 
      since_last_divide ++; 
      if (status) status->update('.');
      if (VERBOSE > 6) fprintf(stdlog, "\n > Adaptive step #%d(of %d): T=%g(%g) dt=%.3g", 
                               i, between_checks, Ca->Time, Gates->Time, dt);  
      if ( rvalue || ( Ca->Time >= T - 2*dt && Ca->Time + dt < T) ) break; // force stability check
    }  // for loop between checks                                          // at end of run time 

    buf_dt = Ca->Time - Gates->Time + 0.5 * dt;
    Buffers->Step(BufStep, Next, *Ca, buf_dt, T);
    Gates->RungeKuttaAdaptive(buf_dt, m_ODEaccuracy, 0, ODEstatus, T);
    if (rvalue) { Plots->draw_all(); if (status) delete status; m_dt0 = dt; return; } // procedure return point 

    Ca->Step( CaStep, Next2, *Buffers, 0.75 * dt, -1);  // check accuracy
    Next2.Step(CaStep, Next, *Buffers, 0.25 * dt, -1);
    two = Next.norm();
    Ca->Step( CaStep, Next, *Buffers, dt, -1);
    one = Next.norm();
    error = (two + one) == 0.0 ? 0 : fabs( 2 * (two - one) / (two + one) );

    if (VERBOSE > 6) 
      fprintf(stdlog,"\n > accuracy check after %d steps: T=%g dt=%g error=%g ", 
              since_last_divide, Ca->Time, dt, error);
    
    if ( isnan(error) || error > 5 * m_accuracy ) { // max error exceeded
       if (status)  status->update('<'); 
       if (VERBOSE > 6) {
         fprintf(stdlog,"\n*** max tolerance exceeded (err = %.2g%% > %.2g%%) \n", error*100, 5*m_accuracy*100);
         fprintf(stdlog,"\n >  Stepping back by %.3gms (to time %.6g) and halving time step to %.3ems\n", 
                     Ca->Time-oldCa.Time, oldCa.Time, dt * 0.5);
       }
       *Ca = oldCa; *Buffers = oldBuf; Gates->recoverState(); Gates->Evaluate();
       dt = (old_dt *= 0.5);
       if (dt < 1e-16) throw ERR_NAN_DT_TINY_ERR;
       Plots->draw_all();
       total_backsteps++;
       between_checks = since_last_divide = 0;
       flag = 0;
       }  
    else if ( error > m_accuracy ) { // halving the time step
       Next2.Time = Ca->Time;
       *Ca = Next2;
       Ca->Time += 0.75 * dt;
       Plots->draw_all();
       dt /= 2;
       if ( error > 2 * m_accuracy) between_checks = m_steps; 
       else between_checks = (2 * since_last_divide) / 3;
       if (status)  status->update('|'); 
       if (VERBOSE > 6) 
         fprintf(stdlog, "\n# dt halved @ %d steps: t=%gms, dt=%.2ems, error=%.2e \n", 
	         since_last_divide, Ca->Time, dt, error);
       since_last_divide = 0;
       flag = 0;
       } 
    else  { // accuracy within limits (error < eps), time step unchanged
       if (!flag) { between_checks = m_steps; flag = 1; }
       if ( error < 0.8 * m_accuracy )  between_checks++;  else between_checks--;
       //between_checks = int( pow( m_accuracy/error, 0.2) / m_dtStretch + 0.5 );
       *Ca = Next;
       Ca->Time += dt;
       oldCa = *Ca;        oldBuf = *Buffers;   // store variables to recover 
       Gates->saveState(); old_dt = dt;         // when error is exceeded
       Plots->draw_all();
       if (status)  status->update('+');
       if (VERBOSE > 6) 
         fprintf(stdlog, "\n> Error within bounds: new time = %lg, one=%lg two=%lg error=%lg\n",Ca->Time,one,two,error);
       }
    } // while 1 ************************************************************
}

//**************************************************************************
//                                R U N                
//**************************************************************************

void SimulationObj::Run() {

  bool adaptive;
  double t, T = 0.0;
  Plots->draw_all();

  for (int i = 1; i <= Params->token_count("Run"); i++)  {
      
     getRun( *Params, i, &adaptive, m_PDEscheme, &t );
     int p = Params->token_index("current", i) + 1;
     if ( equal((*Params)[p], ASSIGN_TOKEN) ) p++;
     ExpressionObj Current(*Params, p, "Bad expression for the current", this, 0, &(Ca->Time), "t");

     Ca->current = &Current;
     T += t;
   
     char *method = StrCpy( getMethod(m_PDEscheme, CaStep, BufStep ) );

       if ( adaptive ) {
          getRun( *Params, i, 0, 0, 0, &m_dt0, &m_accuracy, &m_dtStretch, &m_ODEaccuracy);
          if (VERBOSE) {
            fprintf(stdlog,"\n\n#### Running adaptive %s scheme for %g ms, current = ", method, t);
            Current.print(stdlog);  // make it crash here if expression for current is not valid
            fprintf(stdlog, "\n > dt0=%gms, accuracy=%g, dtStretch=%g, ODEaccuracy=%g \n",
                            m_dt0, m_accuracy, m_dtStretch, m_ODEaccuracy);
	  }
          Adaptive(T);
	}
        else {  // non-adaptive method
          getRun( *Params, i, 0, 0, 0, &m_dt0, &m_ODEaccuracy);
          int Tsteps = int(t / m_dt0 + 0.5);
          if (VERBOSE) {
            fprintf(stdlog,"\n#### Running non-adaptive %s scheme for %g ms, current = ", method, t);
            Current.print(stdlog); fflush(stdlog);
            Current.Evaluate();  // make it crash here is expression for current is not valid
            fprintf(stdlog,"[I]\n > time step = %g ms, number of iterations = %d\n", m_dt0, Tsteps);
          }
          FixedTimeStep(T, Tsteps);
	}
        Plots->redraw_all();
        delete [] method;

        if (VERBOSE) {
          double exper = calcium_gain(Ca, Buffers);
          double theor = *(this->ResolveID("Charge"));
          CHARGE_LOSS = ( exper == theor ) ? 0.0 : 100.0 * (1 - (exper + 1e-20) / (theor + 1e-20) );
          fprintf(stdlog," ## charge loss = %.4g%% (net charge = %g, integrated flux = %g) \n", 
                  CHARGE_LOSS, exper, theor );
	}
        
  } 

  if (VERBOSE) fprintf(stdlog, "\n### Simulation completed: total time = %g ms\n",T );
}
//**************************************************************************
//                           O D E   S O L V E R               
//**************************************************************************

ODESimulationObj::ODESimulationObj(TokenString &TS) : SimulationObj() {

     long pos;
     char temp[512];

     totalSimTime = get_sim_time(TS);
     Params = &TS;
     Gates   = new KineticObj(this);
     if ( TS.token_count("Import", &pos) ) Import( Params->line_string( pos + 1, temp ) );
     Plots   = new PlotArray(*this);
     Plots->redraw_all();

     m_dt0 = 1.0;
     m_ODEaccuracy = 1e-6;
     Params->get_param("ODE.dt0", &m_dt0);
     Params->get_param("ODE.accuracy", &m_ODEaccuracy);

  }                                 

//*******************************************************************************************

void ODESimulationObj::Run()  {
  double time = 0;

  RunStatusString *rss = 0;
  if (VERBOSE > 2) rss = new RunStatusString(30, "time", &(Gates->Time), 0);
  else if (VERBOSE) rss = new RunStatusString(1, "time", &(Gates->Time), 0);

  fprintf(stdlog,"\n\n#### Running the ODEs: total simulation time = %g ms\n", totalSimTime);

  for (int i = 1; i <= Params->token_count("Run"); i++) {
    
    getRun(*Params, i, 0, 0, &time, &m_dt0, &m_ODEaccuracy);

    fprintf(stdlog," ### ODE Run #%d: time = %g ms, accuracy = %g, initial time-step = %g ms\n", 
            i, time, m_ODEaccuracy, m_dt0);

    if (rss) rss->reset( Gates->Time + time );
    Gates->RungeKuttaAdaptive(time, m_ODEaccuracy, Plots, rss, totalSimTime, m_dt0);
    if (VERBOSE) fprintf(stdlog,"\n");
    Plots->redraw_all();
  }

  if (rss) delete rss;
}
//*******************************************************************************************************
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//*******************************************************************************************************
